import { CalculatorType } from "../compute-node";
import { Nub } from "../nub";
import { ParamCalculability, ParamFamily } from "../param/param-definition";
import { ParamValueMode } from "../param/param-value-mode";
import { Result } from "../util/result";
import { MacrorugoParams } from "./macrorugo_params";

export { MacrorugoParams };

export enum MacroRugoFlowType {
    EMERGENT,
    QUASI_EMERGENT,
    SUBMERGED
}

export class MacroRugo extends Nub {

    private static readonly g = 9.81;

    /** nu: water kinematic viscosity  */
    private static readonly nu = 1E-6;
    // Water at 20 °C has a kinematic viscosity of about 10−6 m2·s−1
    // (https://en.wikipedia.org/wiki/Viscosity#Kinematic_viscosity,_%CE%BD)

    /** Ratio between the width (perpendicular to flow) and the lenght (parallel to flow) of a cell (-) */
    private static readonly fracAxAy = 1;

    /** Limit between emergent and submerged flow */
    private static readonly limitSubmerg = 1.1;

    /** Rugosité de fond (m) */
    private ks: number;

    /** Averaged velocity (m.s-1) */

    /** Velocity at the bed (m.s-1) */
    private u0: number;

    /** Discharge (m3/s) */
    private Q: number;

    private _cache: { [key: string]: number };

    constructor(prms: MacrorugoParams, dbg: boolean = false) {
        super(prms, dbg);
        this._cache = {};
        this._calcType = CalculatorType.MacroRugo;
    }

    /**
     * paramètres castés au bon type
     */
    get prms(): MacrorugoParams {
        return this._prms as MacrorugoParams;
    }

    /**
     * Calcul du débit total, de la cote amont ou aval ou d'un paramètre d'une structure
     * @param sVarCalc Nom du paramètre à calculer
     * @param rInit Valeur initiale
     */
    public Calc(sVarCalc: string, rInit?: number): Result {
        // Si on ne force pas à CALCUL, les tests à la 5e décimale ne passent plus (macrorugo.spec.ts)
        const originalValueMode = this.getParameter(sVarCalc).valueMode;
        this.getParameter(sVarCalc).valueMode = ParamValueMode.CALCUL;

        const r: Result = super.Calc(sVarCalc, rInit);
        // Ajout des résultats complémentaires
        // Cote de fond aval
        r.extraResults.ZF2 = this.prms.ZF1.v - this.prms.If.v * this.prms.L.v;
        // Vitesse débitante
        r.extraResults.Vdeb = this.V(this.prms.Q) / this.prms.B.v / this.prms.Y.v;
        // Froude
        const vg: number = r.extraResults.Vdeb / (1 - Math.sqrt(MacroRugo.fracAxAy * this.prms.C.v));
        r.extraResults.Fr = vg / Math.sqrt(MacroRugo.g * this.prms.Y.v);
        // Vitesse maximale
        const cc = 0.4 * this.prms.Cd0.v + 0.7;
        r.extraResults.Vmax = vg * Math.min(
            cc / (1 - Math.pow(r.extraResults.Fr, 2) / 4),
            Math.pow(r.extraResults.Fr, -2 / 3)
        );
        // Puissance dissipée
        r.extraResults.PV = 1000 * MacroRugo.g * this.V(this.prms.Q) / this.prms.B.v * this.prms.If.v;
        // Type d'écoulement
        if (this.prms.Y.v / this.prms.PBH.v < 1) {
            r.extraResults.ENUM_MacroRugoFlowType = MacroRugoFlowType.EMERGENT;
        } else if (this.prms.Y.v / this.prms.PBH.v < MacroRugo.limitSubmerg) {
            r.extraResults.ENUM_MacroRugoFlowType = MacroRugoFlowType.QUASI_EMERGENT;
        } else {
            r.extraResults.ENUM_MacroRugoFlowType = MacroRugoFlowType.SUBMERGED;
        }
        // Vitesse et débit du guide technique
        let cQ: [number, number, number, number];
        let cV: [number, number, number];
        let hdk: number;
        if (r.extraResults.ENUM_MacroRugoFlowType === MacroRugoFlowType.SUBMERGED) {
            cQ = [0.955, 2.282, 0.466, -0.23];
            hdk = this.prms.PBH.v;
        } else {
            hdk = this.prms.PBD.v;
            if (Math.abs(this.prms.Cd0.v - 2) < 0.05) {
                cQ = [0.648, 1.084, 0.56, -0.456];
                cV = [3.35, 0.27, 0.53];
            } else {
                cQ = [0.815, 1.45, 0.557, -0.456];
                cV = [4.54, 0.32, 0.56];
            }
        }
        r.extraResults.Q_GuideTech = cQ[0] * Math.pow(this.prms.Y.v / hdk, cQ[1]) *
            Math.pow(this.prms.If.v, cQ[2]) * Math.pow(this.prms.C.v, cQ[3]) *
            Math.sqrt(MacroRugo.g * hdk) * hdk * this.prms.B.v;
        if (r.extraResults.ENUM_MacroRugoFlowType !== MacroRugoFlowType.SUBMERGED) {
            r.extraResults.V_GuideTech = cV[0] * Math.pow(this.prms.Y.v / this.prms.PBD.v, cV[1]) *
                Math.pow(this.prms.If.v, cV[2]) * Math.sqrt(MacroRugo.g * this.prms.PBD.v);
        }

        // mode d'origine du paramètre; voir le haut de la fonction
        this.getParameter(sVarCalc).valueMode = originalValueMode;

        return r;
    }

    public Equation(sVarCalc: string): Result {
        const Q = uniroot(this.resolveQ, this, 0, 1E7) * this.prms.B.v;
        return new Result(Q, this);
    }

    /**
     * paramétrage de la calculabilité des paramètres
     */
    protected setParametersCalculability() {
        this.prms.ZF1.calculability = ParamCalculability.FREE;
        this.prms.L.calculability = ParamCalculability.FREE;
        this.prms.Ks.calculability = ParamCalculability.FREE;
        this.prms.B.calculability = ParamCalculability.DICHO;
        this.prms.If.calculability = ParamCalculability.DICHO;
        this.prms.Q.calculability = ParamCalculability.EQUATION;
        this.prms.Y.calculability = ParamCalculability.DICHO;
        this.prms.C.calculability = ParamCalculability.DICHO;
        this.prms.PBD.calculability = ParamCalculability.FREE;
        this.prms.PBH.calculability = ParamCalculability.FREE;
        this.prms.Cd0.calculability = ParamCalculability.FREE;
    }

    protected setExtraResultsFamilies() {
        this._extraResultsFamilies = {
            ZF2: ParamFamily.ELEVATIONS,
            Q_GuideTech: ParamFamily.FLOWS
        };
    }

    /**
     * Equation from Cassan, L., Laurens, P., 2016. Design of emergent and submerged rock-ramp fish passes.
     * Knowledge & Management of Aquatic Ecosystems 45.
     * @param sVarCalc Variable à calculer
     */
    private resolveQ(this: MacroRugo, Q: number): number {
        this.Q = Q;
        // Reset cached variables depending on Q (or not...)
        this._cache = {};
        /** Tirant d'eau (m) */
        const h: number = this.prms.Y.v;

        /** Concentration de blocs (-) */
        const C: number = this.prms.C.v;
        /** Paramètre de bloc : Diamètre (m) */
        const D: number = this.prms.PBD.v;
        /** Paramètre de bloc : Hauteur (m) */
        const k: number = this.prms.PBH.v;
        /** adimensional water depth */
        const hstar: number = h / k;
        /** Slope (m/m) */
        const S: number = this.prms.If.v;
        /** Rugosity (m) */
        const ks = this.prms.Ks.v;
        /** Accélération gravité (m/s²) */
        const g = MacroRugo.g;
        /** Constante von Karman */
        const kappa = 0.41;

        if (hstar > MacroRugo.limitSubmerg) {
            // Submerged conditions

            /** Velocity at the bed §2.3.2 Cassan et al., 2016 */
            const cf = 2 / Math.pow(5.1 * Math.log10(k / ks) + 6, 2);
            this.u0 = Math.sqrt(k * 2 * g * S * this.R
                / (this.Cd * C * k / D + cf * this.R));
            /** turbulent length scale (m) within the blocks layer (alpha_t) */
            const alpha = uniroot(this.resolveAlpha_t, this, 1E-6, 100);
            /** averaged velocity at the top of blocks (m.s-1) */
            const uk = this.calcUz(alpha);
            /** Equation (13) Cassan et al., 2016 */
            const d = k - 1 / kappa * alpha * uk / this.ustar;
            /** Equation (14) Cassan et al., 2016 */
            const z0 = (k - d) * Math.exp(- kappa * uk / this.ustar);
            /** Integral of Equation (12) Cassan et al., 2016 */
            // tslint:disable-next-line:variable-name
            const Qsup = this.ustar / kappa * (
                (h - d) * (Math.log((h - d) / z0) - 1)
                - ((k - d) * (Math.log((k - d) / z0) - 1))
            );

            // calcul intégrale dans la canopée----
            // tslint:disable-next-line:variable-name
            let Qinf: number = this.u0;
            let u = 0;
            let uOld: number;
            const step = 0.01;
            const zMax = 1 + step / 2;
            for (let z = step; z < zMax; z += step) {
                uOld = u;
                u = this.calcUz(alpha, z);
                Qinf += (uOld + u) ;
            }
            Qinf = Qinf / 2 * step * k;

            // Calcul de u moyen
            return this.U0 - (Qinf + Qsup) / h;

        } else {
            // Emergent conditions

            // Resolve equation (4) Cassan et al., 2016
            return this.resolveEmergent();
        }
    }

    /**
     * Averaged velocity (m.s-1)
     */
    private get U0(): number {
        if (this._cache.U0 === undefined) {
            this._cache.U0 = this.Q / this.prms.B.v / this.prms.Y.v;
        }
        return this._cache.U0;
    }

    private get CdChD(): number {
        if (this._cache.CdChD === undefined) {
            this._cache.CdChD = this.Cd * this.prms.C.v * this.prms.PBH.v / this.prms.PBD.v;
        }
        return this._cache.CdChD;
    }

    /**
     * sigma ratio between the block area in the x, y plane and D2
     */
    private get sigma(): number {
        if (this._cache.sigma === undefined) {
            if (this.prms.Cd0.v === 2) {
                this._cache.sigma = 1;
            } else {
                this._cache.sigma = Math.PI / 4;
            }
        }
        return this._cache.sigma;
    }

    private get R(): number {
        if (this._cache.R === undefined) {
            this._cache.R = (1 - this.sigma * this.prms.C.v);
        }
        return this._cache.R;
    }

    /**
     * Bed friction coefficient Equation (3) (Cassan et al., 2016)
     */
    private calcCf(): number {

        if (this.prms.Ks.v < 1E-9) {
            // Between Eq (8) and (9) (Cassan et al., 2016)
            const reynolds = this.U0 * this.prms.Y.v / MacroRugo.nu;
            return 0.3164 / 4. * Math.pow(reynolds, -0.25);
        } else {
            // Equation (3) (Cassan et al., 2016)
            return 2 / Math.pow(5.1 * Math.log10(this.prms.Y.v / this.prms.Ks.v - 1) + 6, 2);
        }
    }

    /**
     * Calculation of Cd : drag coefficient of a block under the actual flow conditions
     */
    private get Cd(): number {
        return this.prms.Cd0.v * (1 + 0.4 / Math.pow(this.prms.Y.v / this.prms.PBD.v, 2)) * this.fFr;
    }

    /**
     * Calcul de Beta force ratio between drag and turbulent stress (Cassan et al. 2016 eq(8))
     * \Beta = (k / alpha_t) (C_d C k / D) / (1 - \sigma C)
     * @param alpha \alpha_t turbulent length scale (m) within the blocks layer
     */
    private calcBeta(alpha: number): number {
        return Math.sqrt(this.prms.PBH.v * this.CdChD / alpha / this.R);
    }

    /**
     * Averaged velocity at a given vertical position (m.s-1)
     * @param alpha turbulent length scale (m) within the blocks layer
     * @param z dimensionless vertical position z / k
     */
    private calcUz(alpha: number, z: number = 1): number {
        const beta = this.calcBeta(alpha);
        // Equation (9) Cassan et al., 2016
        return this.u0 * Math.sqrt(
            beta * (this.prms.Y.v / this.prms.PBH.v - 1) * Math.sinh(beta * z) / Math.cosh(beta) + 1
        );
    }

    private get ustar(): number {
        if (this._cache.ustar === undefined) {
            this._cache.ustar = Math.sqrt(MacroRugo.g * this.prms.If.v * (this.prms.Y.v - this.prms.PBH.v));
        }
        return this._cache.ustar;
    }

    private resolveAlpha_t(alpha: number): number {
        /** s: minimum distance between blocks */
        const s = this.prms.PBD.v * (1 / Math.sqrt(this.prms.C.v) - 1);
        /** Equation(11) Cassan et al., 2016 */
        const l0 = Math.min(s, 0.15 * this.prms.PBH.v);
        // Equation(10) Cassan et al., 2016
        return alpha * this.calcUz(alpha) - l0 * this.ustar;
    }

    private resolveEmergent(): number {
        const g = MacroRugo.g;

        const alpha = 1 - Math.pow(1 * this.prms.C.v, 0.5) - 0.5 * this.sigma * this.prms.C.v;
        // tslint:disable-next-line:variable-name
        const Cd = this.prms.Cd0.v * (1 + 1 / Math.pow(this.prms.Y.v / this.prms.PBD.v, 2)) * this.fFr;

        /** N from Cassan 2016 eq(2) et Cassan 2014 eq(12) */
        const N = (alpha * this.calcCf()) / (this.prms.Y.v / this.prms.PBD.v * Cd * this.prms.C.v);

        return this.U0 - Math.sqrt(
            2 * MacroRugo.g * this.prms.If.v * this.prms.PBD.v *
            (1 - this.sigma * this.prms.C.v) / (Cd * this.prms.C.v * (1 + N))
        );
    }

    /**
     * Froude correction function
     */
    private get fFr(): number {
        if (this._cache.fFr === undefined) {
            // tslint:disable-next-line:variable-name
            const Fr = this.U0 /
                (1 - Math.sqrt(MacroRugo.fracAxAy * this.prms.C.v)) /
                Math.sqrt(MacroRugo.g * this.prms.Y.v);

            /** Interpolation linéaire entre le bloc rond (Cd0=1) et le carré (Cd0=2) */
            const r = 0.4 * this.prms.Cd0.v + 0.7;

            if (Fr < 1.3) {
                this._cache.fFr = Math.pow(Math.min(r / (1 - Math.pow(Fr, 2) / 4), Math.pow(Fr, -2 / 3)), 2);
            } else {
                this._cache.fFr = Math.pow(Fr, -4 / 3);
            }
        }
        return this._cache.fFr;
    }
}

/**
 * Searches the interval from <tt>lowerLimit</tt> to <tt>upperLimit</tt>
 * for a root (i.e., zero) of the function <tt>func</tt> with respect to
 * its first argument using Brent's method root-finding algorithm.
 *
 * Translated from zeroin.c in http://www.netlib.org/c/brent.shar.
 *
 * Copyright (c) 2012 Borgar Thorsteinsson <borgar@borgar.net>
 * MIT License, http://www.opensource.org/licenses/mit-license.php
 *
 * @param {function} func function for which the root is sought.
 * @param {number} lowerlimit the lower point of the interval to be searched.
 * @param {number} upperlimit the upper point of the interval to be searched.
 * @param {number} errorTol the desired accuracy (convergence tolerance).
 * @param {number} maxIter the maximum number of iterations.
 * @returns an estimate for the root within accuracy.
 *
 */
function uniroot<T>(func: (param: number) => number, thisArg: T, lowerLimit: number, upperLimit: number,
                    errorTol: number = 0, maxIter: number = 1000
) {
    let a = lowerLimit;
    let b = upperLimit;
    let c = a;
    let fa = func.call(thisArg, a);
    let fb = func.call(thisArg, b);
    let fc = fa;
    let tolAct;   // Actual tolerance
    let newStep;  // Step at this iteration
    let prevStep; // Distance from the last but one to the last approximation
    let p;        // Interpolation step is calculated in the form p/q; division is delayed until the last moment
    let q;

    while (maxIter-- > 0) {

        prevStep = b - a;

        if (Math.abs(fc) < Math.abs(fb)) {
            // Swap data for b to be the best approximation
            a = b, b = c, c = a;
            fa = fb, fb = fc, fc = fa;
        }

        tolAct = 1e-15 * Math.abs(b) + errorTol / 2;
        newStep = (c - b) / 2;

        if (Math.abs(newStep) <= tolAct || fb === 0) {
            return b; // Acceptable approx. is found
        }

        // Decide if the interpolation can be tried
        if (Math.abs(prevStep) >= tolAct && Math.abs(fa) > Math.abs(fb)) {
            // If prev_step was large enough and was in true direction, Interpolatiom may be tried
            let t1;
            let cb;
            let t2;
            cb = c - b;
            if (a === c) { // If we have only two distinct points linear interpolation can only be applied
                t1 = fb / fa;
                p = cb * t1;
                q = 1.0 - t1;
            } else { // Quadric inverse interpolation
                q = fa / fc, t1 = fb / fc, t2 = fb / fa;
                p = t2 * (cb * q * (q - t1) - (b - a) * (t1 - 1));
                q = (q - 1) * (t1 - 1) * (t2 - 1);
            }

            if (p > 0) {
                q = -q;  // p was calculated with the opposite sign; make p positive
            } else {
                p = -p;  // and assign possible minus to q
            }

            if (p < (0.75 * cb * q - Math.abs(tolAct * q) / 2) &&
                p < Math.abs(prevStep * q / 2)) {
                // If (b + p / q) falls in [b,c] and isn't too large it is accepted
                newStep = p / q;
            }

            // If p/q is too large then the bissection procedure can reduce [b,c] range to more extent
        }

        if (Math.abs(newStep) < tolAct) { // Adjust the step to be not less than tolerance
            newStep = (newStep > 0) ? tolAct : -tolAct;
        }

        a = b, fa = fb;     // Save the previous approx.
        b += newStep, fb = func.call(thisArg, b);  // Do step to a new approxim.

        if ((fb > 0 && fc > 0) || (fb < 0 && fc < 0)) {
            c = a, fc = fa; // Adjust c for it to have a sign opposite to that of b
        }
    }
    return undefined;

}
