import { ParamDefinition, ParamFamily } from "../param/param-definition";
import { ParamDomainValue } from "../param/param-domain";
import { ParamsEquation } from "../param/params-equation";

export class MacrorugoParams extends ParamsEquation {
    [key: string]: any; // pour pouvoir faire this['methode']();

    /** Cote de fond amont (m) */
    private _ZF1: ParamDefinition;

    /** Longueur (m) */
    private _L: ParamDefinition;

    /** Largeur (m) */
    private _B: ParamDefinition;

    /** Pente (m/m) */
    private _If: ParamDefinition;

    /** Débit (m3/s) */
    private _Q: ParamDefinition;

    /** Tirant d'eau (m) */
    private _Y: ParamDefinition;

    /** Rugosité de fond (m) */
    private _Ks: ParamDefinition;

    /** Concentration de blocs (-) */
    private _C: ParamDefinition;

    /** Paramètre de bloc : Diamètre (m) */
    private _PBD: ParamDefinition;

    /** Paramètre de bloc : Hauteur (m) */
    private _PBH: ParamDefinition;

    /** Paramètre de bloc : Forme (1 pour rond, 2 pour carré) */
    private _Cd0: ParamDefinition;

    /**
     *
     * @param rZF1 Cote de fond amont (m)
     * @param rL Longueur (m)
     * @param rB Largeur (m)
     * @param rIf Pente (m/m)
     * @param rQ Débit (m3/s)
     * @param rY Tirant d'eau (m)
     * @param rRF Rugosité de fond (m)
     * @param rCB Concentration de blocs (m)
     * @param rPBD Paramètre de bloc : Diamètre (m)
     * @param rPBH Paramètre de bloc : Hauteur (m)
     * @param rCd0 Paramètre de bloc : Forme (1 pour rond, 2 pour carré)
     */
    constructor(rZF1: number, rL: number, rB: number, rIf: number, rQ: number,
                rY: number, rRF: number, rCB: number, rPBD: number, rPBH: number, rCd0: number) {
        super();

        this._ZF1 = new ParamDefinition(this, "ZF1", ParamDomainValue.POS, rZF1, ParamFamily.ELEVATIONS);
        this.addParamDefinition(this._ZF1);

        this._L = new ParamDefinition(this, "L", ParamDomainValue.POS, rL, ParamFamily.LENGTHS);
        this.addParamDefinition(this._L);

        this._B = new ParamDefinition(this, "B", ParamDomainValue.POS, rB, ParamFamily.WIDTHS);
        this.addParamDefinition(this._B);

        this._If = new ParamDefinition(this, "If", ParamDomainValue.POS, rIf, ParamFamily.SLOPES);
        this.addParamDefinition(this._If);

        this._Q = new ParamDefinition(this, "Q", ParamDomainValue.POS, rQ, ParamFamily.FLOWS);
        this.addParamDefinition(this._Q);

        this._Y = new ParamDefinition(this, "Y", ParamDomainValue.POS, rY, ParamFamily.HEIGHTS);
        this.addParamDefinition(this._Y);

        this._Ks = new ParamDefinition(this, "Ks", ParamDomainValue.POS_NULL, rRF);
        this.addParamDefinition(this._Ks);

        this._C = new ParamDefinition(this, "C", ParamDomainValue.POS, rCB);
        this.addParamDefinition(this._C);

        this._PBD = new ParamDefinition(this, "PBD", ParamDomainValue.POS, rPBD);
        this.addParamDefinition(this._PBD);

        this._PBH = new ParamDefinition(this, "PBH", ParamDomainValue.POS, rPBH, ParamFamily.HEIGHTS);
        this.addParamDefinition(this._PBH);

        this._Cd0 = new ParamDefinition(this, "Cd0", ParamDomainValue.POS, rCd0);
        this.addParamDefinition(this._Cd0);

    }

    /**
     * Cote de fond amont (m)
     * @return {ParamDefinition}
     */
    public get ZF1(): ParamDefinition {
        return this._ZF1;
    }

    /**
     * Longueur (m)
     * @return {ParamDefinition}
     */
    public get L(): ParamDefinition {
        return this._L;
    }

    /**
     * Getter B
     * @return {ParamDefinition}
     */
    public get B(): ParamDefinition {
        return this._B;
    }

    /**
     * Pente (m/m)
     * @return {ParamDefinition}
     */
    public get If(): ParamDefinition {
        return this._If;
    }

    /**
     * Débit (m3/s)
     * @return {ParamDefinition}
     */
    public get Q(): ParamDefinition {
        return this._Q;
    }

    /**
     * Tirant d'eau (m)
     * @return {ParamDefinition}
     */
    public get Y(): ParamDefinition {
        return this._Y;
    }

    /**
     * Rugosité de fond (m)
     * @return {ParamDefinition}
     */
    public get Ks(): ParamDefinition {
        return this._Ks;
    }

    /**
     * Concentration de blocs (-)
     * @return {ParamDefinition}
     */
    public get C(): ParamDefinition {
        return this._C;
    }

    /**
     * Paramètre de bloc : Diamètre (m)
     * @return {ParamDefinition}
     */
    public get PBD(): ParamDefinition {
        return this._PBD;
    }

    /**
     * Paramètre de bloc : Hauteur (m)
     * @return {ParamDefinition}
     */
    public get PBH(): ParamDefinition {
        return this._PBH;
    }

    /**
     * Paramètre de bloc : Forme (1 pour rond, 2 pour carré)
     * @return {ParamDefinition}
     */
    public get Cd0(): ParamDefinition {
        return this._Cd0;
    }

}
