import { ParamCalculability, ParamDefinition, ParamFamily} from "../param/param-definition";
import { ParamDomainValue } from "../param/param-domain";
import { Result } from "../util/result";

import { CalculatorType, Session } from "../index";
import { Cloisons, CloisonsParams } from "../structure/cloisons";
import { Pab } from "./pab";

class PabCloisonsParams extends CloisonsParams {

    /** Débit d'attrait d'un bassin (m3/s) */
    public _QA: ParamDefinition;

    /** Model of cloison on which this pabCloison is based */
    private _modelCloisonsParams: CloisonsParams;

    /**
     * Paramètres communs à toutes les équations de structure
     * @param rQ Débit total (m3/s)
     * @param rZ1 Cote de l'eau amont (m)
     * @param rLB Longueur des bassins (m)
     * @param rBB Largeur des bassins (m)
     * @param rPB Profondeur moyenne (m)
     * @param rDH Hauteur de chute (m)
     */
    constructor(modelCloisonsParams?: CloisonsParams, rQA: number = 0) {
        super(1, 1, 1, 1, 1, 1); // overwritten by init() below
        this.modelCloisonsParams = modelCloisonsParams;

        // Débit d'attrait
        this._QA = new ParamDefinition(this, "QA", ParamDomainValue.POS_NULL, rQA, ParamFamily.FLOWS);
        this.addParamDefinition(this._QA);
    }

    public get QA() {
        return this._QA;
    }

    /**
     * Record pointer to the cloison model for future updating of local parameters
     */
    set modelCloisonsParams(modelCloisonsParams: CloisonsParams) {
        this._modelCloisonsParams = modelCloisonsParams;
    }

    /**
     * Update Current values and sandobx values of parameter from cloison model
     */
    public setCurrentValuesFromModel() {
        if (this._modelCloisonsParams !== undefined) {
            for (const p of this._modelCloisonsParams) {
                this._paramMap[p.symbol].currentValue = p.currentValue;
                this._paramMap[p.symbol].v = p.currentValue;
            }
        }
    }

}

// tslint:disable-next-line:max-classes-per-file
export class PabCloisons extends Cloisons {

    /**
     * paramètres castés au bon type
     */
    get prms(): PabCloisonsParams {
        return this._prms as PabCloisonsParams;
    }

    /**
     * Calculation of upstream water depth
     */
    get Yam(): number {
        // TODO: ajouter l'option radier horizontal
        return this.prms.PB.currentValue + this.prms.DH.currentValue / 2;
    }

    public parent: Pab;

    constructor(modelCloisons: Cloisons, rQA: number = 0, dbg: boolean = false) {
        super(new PabCloisonsParams(), dbg); // model is set below
        this.initModelCloisons(modelCloisons);
        this._calcType = CalculatorType.PabCloisons;
    }

    public Calc(sVarCalc: string, rInit?: number): Result {
        const r: Result = super.Calc(sVarCalc, rInit);
        const p = this.getParamValuesAfterCalc(sVarCalc, r);
        r.extraResults.ZRAM = p.Z1 - p.PB;
        return r;
    }

    /**
     * Update crest elevations from upstream apron elevation
     */
    public updateZDV(rZR: number) {
        for (const st of this.structures) {
            if (st.prms.ZDV.calculability !== ParamCalculability.NONE) {
                st.prms.ZDV.v = st.prms.ZDV.currentValue + rZR - (this.prms.Z1.currentValue - this.Yam);
            }
        }
    }

    /**
     * Sets modelCloisons as the current model (set pointers to parameters and structure array)
     */
    public initModelCloisons(modelCloisons: Cloisons) {
        this.prms.modelCloisonsParams = modelCloisons ? modelCloisons.prms : undefined;
        this._children = modelCloisons ? modelCloisons.structures : [];
    }

    /**
     * Finds the Cloisons targetted by modelUid and defines it as the current model
     */
    public setModel(modelUid: string) {
        this.properties.setPropValue("modeleCloisons", modelUid);
        const cloisons = (Session.getInstance().findNubByUid(modelUid) as Cloisons);
        if (cloisons) {
            this.initModelCloisons(cloisons);
        } /* else {
            console.error("PabCloisons.setModel : cannot find target Cloisons");
        } */
    }

    /**
     * Returns an object representation of the Nub's current state
     */
    public objectRepresentation() {
        const ret: any = {
            uid: this.uid,
            props: this.properties.props,
            parameters: []
        };

        // ! do not iterate over local parameters (would copy the parameters of the Cloisons used as model),
        // just serialise QA
        ret.parameters.push(this.prms.QA.objectRepresentation());

        // ! do not iterate over children Nubs (would copy the children of the Cloisons used as model)

        return ret;
    }

    /**
     * paramétrage de la calculabilité des paramètres
     */
    protected setParametersCalculability() {
        super.setParametersCalculability();
        this.prms._QA.calculability = ParamCalculability.FREE;
    }
}
