import { floatDivAndMod } from "../base";
import { CalculatorType } from "../compute-node";
import { Nub } from "../nub";
import { ParamCalculability, ParamDefinition, ParamFamily } from "../param/param-definition";
import { ParamDomainValue } from "../param/param-domain";
import { ParamsEquation } from "../param/params-equation";
import { Message, MessageCode } from "../util/message";
import { Result } from "../util/result";

export class PabNombreParams extends ParamsEquation {
    [key: string]: any; // pour pouvoir faire this['methode']();

    /** Chute totale DHT */
    private _DHT: ParamDefinition;

    /** Nombre de chutes N */
    private _N: ParamDefinition;

    /** Chute entre bassins DH */
    private _DH: ParamDefinition;

    constructor(rDHT: number, rN: number, rDH: number) {
        super();
        this._DHT = new ParamDefinition(this, "DHT", ParamDomainValue.POS, rDHT, ParamFamily.TOTALFALLS);
        this._N = new ParamDefinition(this, "N", ParamDomainValue.POS, rN);
        this._DH = new ParamDefinition(this, "DH", ParamDomainValue.POS, rDH, ParamFamily.BASINFALLS);

        this.addParamDefinition(this._DHT);
        this.addParamDefinition(this._N);
        this.addParamDefinition(this._DH);
    }

    get DHT() {
        return this._DHT;
    }

    get N() {
        return this._N;
    }

    get DH() {
        return this._DH;
    }
}

// tslint:disable-next-line:max-classes-per-file
export class PabNombre extends Nub {
    constructor(prms: PabNombreParams, dbg: boolean = false) {
        super(prms, dbg);
        this._calcType = CalculatorType.PabNombre;
        this._defaultCalculatedParam = prms.N;
        this.resetDefaultCalculatedParam();
    }

    /**
     * paramètres castés au bon type
     */
    get prms(): PabNombreParams {
        return this._prms as PabNombreParams;
    }

    public Equation(sVarCalc: string): Result {
        if (sVarCalc !== "N" && ! Number.isInteger(this.prms.N.v)) {
            const m = new Message(MessageCode.ERROR_PARAMDEF_VALUE_INTEGER);
            m.extraVar.N = this.prms.N.v;
            return new Result(m);
        }

        let v: number;
        let DHR = 0;

        switch (sVarCalc) {
            case "DHT":
                v = this.prms.N.v * this.prms.DH.v;
                break;

            case "N":
                const divAndMod = floatDivAndMod(this.prms.DHT.v, this.prms.DH.v);
                v = divAndMod.q;
                DHR = divAndMod.r;
                break;

            case "DH":
                v = this.prms.DHT.v / this.prms.N.v;
                break;

            default:
                throw new Error("PabNombre.Equation() : invalid variable name " + sVarCalc);
        }

        const r = new Result(v, this);
        r.extraResults.DHR = DHR;

        return r;
    }

    /**
     * paramétrage de la calculabilité des paramètres
     */
    protected setParametersCalculability() {
        this.prms.DHT.calculability = ParamCalculability.EQUATION;
        this.prms.N.calculability = ParamCalculability.EQUATION;
        this.prms.DH.calculability = ParamCalculability.EQUATION;
    }
}
