import { CalculatorType, ComputeNode } from "../compute-node";
import { LinkedValue } from "../linked-value";
import { Nub } from "../nub";
import { acSection } from "../section/section_type";
import { Session } from "../session";
import { Interval } from "../util/interval";
import { Message, MessageCode } from "../util/message";
import { IObservable, Observable, Observer } from "../util/observer";
import { ParamDomain, ParamDomainValue } from "./param-domain";
import { INamedIterableValues, INumberIterator } from "./param-value-iterator";
import { ParamValueMode } from "./param-value-mode";
import { ParamValues } from "./param-values";
import { ParamsEquation } from "./params-equation";

/**
 * Calculabilité du paramètre
 */
export enum ParamCalculability {
    /** paramètre fixé (immuable, cad non modifiable après création) */
    NONE,
    /** paramètre libre (modifiable) */
    FREE,
    /** paramètre calculable analytiquement, par méthode de Newton, ... */
    EQUATION,
    /** paramètre calculable par dichotomie */
    DICHO
}

/**
 * Parameter family: defines linkability with other parameters/results
 */
export enum ParamFamily {
    LENGTHS, // longueur
    WIDTHS, // largeur
    SLOPES, // pente
    HEIGHTS, // profondeur, tirant d'eau
    BASINFALLS, // chute entre bassins
    TOTALFALLS, // chute totale
    ELEVATIONS, // cote
    VOLUMES,
    FLOWS, // débit
    DIAMETERS,
    // SPEEDS // vitesses, seulement des résultats
}

/**
 * Paramètre avec symbole, famille, domaine de définition, calculabilité,
 * pointant éventuellement vers un autre paramètre / résultat
 */
export class ParamDefinition implements INamedIterableValues, IObservable {

    /** le paramètre doit-il être exposé (par ex: affiché par l'interface graphique) ? */
    public visible: boolean = true;

    /** sandbox value used during calculation */
    public v: number;

    /** mode de génération des valeurs : min/max, liste, ... */
    private _valueMode: ParamValueMode;

    /** symbole */
    private _symbol: string;

    /** related parameters groups; indirectly gives access to the Nub using this parameter */
    private _parent: ParamsEquation;

    /** domaine de définition */
    private _domain: ParamDomain;

    /** calculabilité */
    private _calc: ParamCalculability;

    /** valeur(s) prise(s) */
    private _paramValues: ParamValues;

    /** parameters family, for linking */
    private _family: ParamFamily;

    /** pointer to another Parameter / Result, in case of LINK mode */
    private _referencedValue: LinkedValue;

    /** implémentation par délégation de IObservable */
    private _observable: Observable;

    constructor(parent: ParamsEquation, symb: string, d: ParamDomain | ParamDomainValue,
                val?: number, family?: ParamFamily, visible: boolean = true
    ) {
        this._parent = parent;
        this._symbol = symb;
        this._observable = new Observable();
        this._paramValues = new ParamValues();

        // set single value and copy it to currentValue
        this._paramValues.singleValue = val;
        this.v = val;

        this._calc = ParamCalculability.FREE;
        this._family = family;
        this.visible = visible;
        this.valueMode = ParamValueMode.SINGLE;

        if (d instanceof ParamDomain) {
            this._domain = d;
        } else {
            this._domain = new ParamDomain(d as ParamDomainValue);
        }

        this.checkValueAgainstDomain(val);
    }

    /**
     * set parent a-posteriori; used by nghyd when populating forms
     */
    public set parent(parent: ParamsEquation) {
        this._parent = parent;
    }

    /**
     * identifiant unique de la forme
     *   (uid du JalhydObject (en général un Nub) + "_" + symbole du paramètre)
     * ex: 123_Q
     */
    public get uid(): string {
        return this.nubUid + "_" + this._symbol;
    }

    /**
     * pointer to the ComputeNode (usually Nub) that uses the ParamsEquation that
     * uses this ParamDefinition; for Section params, this means the Nub enclosing
     * the Section
     */
    public get parentComputeNode(): ComputeNode {
        let parentCN: ComputeNode;
        if (this._parent) {
            // ComputeNode utilisant le ParamsEquation
            parentCN = this._parent.parent;
            // Section: go up to enclosing Nub if any (might not have any yet when unserializing sessions)
            if (parentCN instanceof acSection && parentCN.parent) {
                parentCN = parentCN.parent;
            }
        } else {
            // fail silently (the story of my life)
            // throw new Error("ParamDefinition.parentComputeNode : parameter has no parent !");
        }
        return parentCN;
    }

    /**
     * Returns the parent compute node as Nub if it is a Nub; returns
     * undefined otherwise
     */
    public get parentNub(): Nub {
        if (this.parentComputeNode && this.parentComputeNode instanceof Nub) {
            return this.parentComputeNode as Nub;
        } else {
            return undefined;
        }
    }

    /**
     * Identifiant unique du Nub parent
     */
    public get nubUid(): string {
        return this.parentComputeNode.uid;
    }

    /**
     * Type de module du Nub parent
     */
    public get nubCalcType(): CalculatorType {
        let parentCalcType: CalculatorType;
        if (this.parentComputeNode instanceof Nub) {
            parentCalcType = this.parentComputeNode.calcType;
        } else {
            throw new Error("ParamDefinition.nubCalcType : parameter has no parent !");
        }
        return parentCalcType;
    }

    public get symbol(): string {
        return this._symbol;
    }

    public get domain(): ParamDomain {
        return this._domain;
    }

    public get interval(): Interval {
        return this._domain.interval;
    }

    public get valueMode(): ParamValueMode {
        return this._valueMode;
    }

    /**
     * Easy setter that propagates the change to other parameters
     * (default behaviour); use setValueMode(..., false) to prevent
     * possible infinite loops
     */
    public set valueMode(newMode: ParamValueMode) {
        this.setValueMode(newMode);
    }

    /**
     * Sets the value mode and asks the Nub to ensure all parameters requirements are met
     * @see Nub::ensureParametersConsistency()
     *
     * If propagateToCalculatedParam is true, any param that goes from CALC mode
     * to any other mode will trigger a reset of the default calculated param at
     * Nub level
     *
     * Propagates modes other than CALC and LINK to the underlying
     * ParamValues (local instance)
     */
    public setValueMode(newMode: ParamValueMode, propagateToCalculatedParam: boolean = true) {
        const oldMode = this.valueMode;

        // ignore idempotent calls
        if (oldMode === newMode) {
            return;
        }

        switch (oldMode) {
            case ParamValueMode.SINGLE:  // ancien état
                switch (newMode) {
                    case ParamValueMode.MINMAX:  // nouvel état
                    case ParamValueMode.LISTE:
                        this.ensureParametersConsistency(this, ParamValueMode.CALCUL);
                        break;

                    case ParamValueMode.CALCUL:  // nouvel état
                        this.ensureParametersConsistency(this, [ ParamValueMode.MINMAX, ParamValueMode.LISTE ]);
                        break;

                    case ParamValueMode.LINK:  // nouvel état
                        if (
                            this.isReferenceDefined()
                            && this.hasMultipleValues
                        ) {
                            this.ensureParametersConsistency(this); // reset existing variable param
                        }
                        break;
                }
                break;

            case ParamValueMode.LISTE:  // ancien état
            case ParamValueMode.MINMAX:
                switch (newMode) {
                    case ParamValueMode.CALCUL:  // nouvel état
                        this.ensureParametersConsistency(this);
                        break;

                    case ParamValueMode.LINK:  // nouvel état
                        if (
                            this.isReferenceDefined()
                            && this.hasMultipleValues
                        ) {
                            this.ensureParametersConsistency(this); // reset existing variable param
                        }
                        break;
                }
                break;

            case ParamValueMode.LINK:  // ancien état
                switch (newMode) {
                    case ParamValueMode.MINMAX:  // nouvel état
                    case ParamValueMode.LISTE:
                        this.ensureParametersConsistency(this, ParamValueMode.CALCUL);
                        break;

                    case ParamValueMode.CALCUL:  // nouvel état
                        this.ensureParametersConsistency(this, [ ParamValueMode.MINMAX, ParamValueMode.LISTE ]);
                        break;
                }
                break;

            case ParamValueMode.CALCUL:  // ancien état
                switch (newMode) {
                    case ParamValueMode.SINGLE:  // nouvel état
                        break;

                    case ParamValueMode.MINMAX:  // nouvel état
                    case ParamValueMode.LISTE:
                        this.ensureParametersConsistency(this);
                        break;

                    case ParamValueMode.LINK:  // nouvel état
                        if (
                            this.isReferenceDefined()
                            && this.hasMultipleValues
                        ) {
                            this.ensureParametersConsistency(this); // reset existing variable param
                        }
                        break;
                }

                if (propagateToCalculatedParam) {
                    // Set default calculated parameter, only if previous CALC param was
                    // manually set to something else than CALC
                    if (this.parentComputeNode && this.parentComputeNode instanceof Nub) {
                        this.parentComputeNode.resetDefaultCalculatedParam(this);
                    }
                }

        }

        // set new mode
        this._valueMode = newMode;

        // synchronise with underlying local ParamValues (for iterator)
        if (
            newMode === ParamValueMode.SINGLE
            || newMode === ParamValueMode.MINMAX
            || newMode === ParamValueMode.LISTE
        ) {
            if (this._paramValues) {
                this._paramValues.valueMode = newMode;
            }
        }
    }

    /**
     * Returns true if this parameter is the calculatedParam of
     * its parent Nub
     */
    public get isCalculated(): boolean {
        if (this.parentNub) {
            return (this.parentNub.calculatedParam === this); // should be the same object
        }
        return false;
    }

    /**
     * Sets this parameter as the one to be calculated
     */
    public setCalculated() {
        if (this.parentNub) {
            this.parentNub.calculatedParam = this;
        }
    }

    /**
     * Current values set associated to this parameter; in LINK mode, points
     * to a remote set of values.
     */
    public get paramValues(): ParamValues {
        if (this.valueMode === ParamValueMode.LINK && this.isReferenceDefined()) {
            return this._referencedValue.getParamValues();
        } else {
            return this._paramValues;
        }
    }

    public get referencedValue() {
        return this._referencedValue;
    }

    public get family() {
        return this._family;
    }

    public undefineFamily() {
        this._family = undefined;
    }

    get calculability(): ParamCalculability {
        if (this._calc === undefined) {
            const e = new Message(MessageCode.ERROR_PARAMDEF_CALC_UNDEFINED);
            e.extraVar.symbol = this.symbol;
            throw e;
        }

        return this._calc;
    }

    set calculability(c: ParamCalculability) {
        this._calc = c;
    }

    /**
     * Returns true if current value (not singleValue !) is defined
     */
    public get hasCurrentValue(): boolean {
        return (this.v !== undefined);
    }

    /**
     * Returns true if held value (not currentValue !) is defined,
     * depending on the value mode
     */
    public get isDefined(): boolean {
        let defined = false;
        switch (this.valueMode) {
            case ParamValueMode.SINGLE:
                defined = (this.singleValue !== undefined);
                break;
            case ParamValueMode.MINMAX:
                defined = (
                    this.paramValues.min !== undefined
                    && this.paramValues.max !== undefined
                    && this.paramValues.step !== undefined
                );
                break;
            case ParamValueMode.LISTE:
                defined = (this.valueList.length > 0 && this.valueList[0] !== undefined);
                break;
            case ParamValueMode.CALCUL:
                if (this.parentNub && this.parentNub.result && this.parentNub.result.nbResultElements > 0) {
                    const res = this.parentNub.result;
                    defined = (res.vCalc !== undefined);
                }
                break;
            case ParamValueMode.LINK:
                defined = this.referencedValue.isDefined();
        }
        return defined;
    }

    // -- values getters / setters; in LINK mode, reads / writes from / to the target values

    public get singleValue(): number {
        this.checkValueMode([ParamValueMode.SINGLE, ParamValueMode.CALCUL, ParamValueMode.LINK]);
        return this.paramValues.singleValue;
    }

    public set singleValue(v: number) {
        this.checkValueMode([ParamValueMode.SINGLE, ParamValueMode.CALCUL, ParamValueMode.LINK]);
        this.paramValues.singleValue = v;
        this.notifyValueModified(this);
    }

    public get min() {
        this.checkValueMode([ParamValueMode.MINMAX, ParamValueMode.LINK]);
        return this.paramValues.min;
    }

    public set min(v: number) {
        this.checkValueMode([ParamValueMode.MINMAX]);
        this.paramValues.min = v;
    }

    public get max() {
        this.checkValueMode([ParamValueMode.MINMAX, ParamValueMode.LINK]);
        return this.paramValues.max;
    }

    public set max(v: number) {
        this.checkValueMode([ParamValueMode.MINMAX]);
        this.paramValues.max = v;
    }

    public get step() {
        this.checkValueMode([ParamValueMode.MINMAX, ParamValueMode.LINK]);
        return this.paramValues.step;
    }

    public set step(v: number) {
        this.checkValueMode([ParamValueMode.MINMAX]);
        this.paramValues.step = v;
    }

    /**
     * Generates a reference step value, given the current (local) values for min / max
     */
    public get stepRefValue(): Interval {
        this.checkValueMode([ParamValueMode.MINMAX]);
        return new Interval(1e-9, this._paramValues.max - this._paramValues.min);
    }

    public get valueList() {
        this.checkValueMode([ParamValueMode.LISTE, ParamValueMode.LINK]);
        return this.paramValues.valueList;
    }

    public set valueList(l: number[]) {
        this.checkValueMode([ParamValueMode.LISTE]);
        this.paramValues.valueList = l;
    }

    public count() {
        return this.paramValues.valuesIterator.count();
    }

    // for INumberiterator interface
    public get currentValue(): number {
        // magically follows links
        return this.paramValues.currentValue;
    }

    public set currentValue(v: number) {
        // magically follows links
        this.paramValues.currentValue = v;
    }

    /**
     * Get single value
     */
    public getValue(): number {
        return this.paramValues.singleValue;
    }

    /**
     * Magic method to define current values into Paramvalues set; defines the mode
     * accordingly by detecting values nature
     */
    public setValues(o: number | any, max?: number, step?: number) {
        this.paramValues.setValues(o, max, step);
        if (typeof (o) === "number") {
            if (max === undefined) {
                this.valueMode = ParamValueMode.SINGLE;
            } else {
                this.valueMode = ParamValueMode.MINMAX;
            }
        } else if (Array.isArray(o)) {
            this.valueMode = ParamValueMode.LISTE;
        } else {
            throw new Error(`ParamValues.setValues() : invalid call`);
        }
    }

    /**
     * Sets the current value of the parameter's own values set, then
     * notifies all observers
     */
    public setValue(val: number, sender?: any) {
        this.checkValueAgainstDomain(val);
        this.paramValues.singleValue = val;
        // this.notifyValueModified(sender); // @TODO why ? currentValue is only used temporarily for calculation
    }

    /**
     * Checks that the current values set are compatible with the value mode
     */
    public checkValueAgainstMode() {
        this.paramValues.check();
    }

    /**
     * Validates the given numeric value against the definition domain; throws
     * an error if value is outside the domain
     */
    public checkValueAgainstDomain(v: number) {
        const sDomain = ParamDomainValue[this._domain.domain];

        switch (this._domain.domain) {
            case ParamDomainValue.ANY:
                break;

            case ParamDomainValue.POS:
                if (v <= 0) {
                    const f = new Message(MessageCode.ERROR_PARAMDEF_VALUE_POS);
                    f.extraVar.symbol = this.symbol;
                    f.extraVar.value = v;
                    throw f;
                }
                break;

            case ParamDomainValue.POS_NULL:
                if (v < 0) {
                    const f = new Message(MessageCode.ERROR_PARAMDEF_VALUE_POSNULL);
                    f.extraVar.symbol = this.symbol;
                    f.extraVar.value = v;
                    throw f;
                }
                break;

            case ParamDomainValue.NOT_NULL:
                if (v === 0) {
                    const f = new Message(MessageCode.ERROR_PARAMDEF_VALUE_NULL);
                    f.extraVar.symbol = this.symbol;
                    throw f;
                }
                break;

            case ParamDomainValue.INTERVAL:
                const min = this._domain.minValue;
                const max = this._domain.maxValue;
                if (v < min || v > max) {
                    const f = new Message(MessageCode.ERROR_PARAMDEF_VALUE_INTERVAL);
                    f.extraVar.symbol = this.symbol;
                    f.extraVar.value = v;
                    f.extraVar.minValue = min;
                    f.extraVar.maxValue = max;
                    throw f;
                }
                break;

            default:
                const e = new Message(MessageCode.ERROR_PARAMDOMAIN_INVALID);
                e.extraVar.symbol = this.symbol;
                e.extraVar.domain = sDomain;
                throw e;
        }
    }

    public checkMin(min: number): boolean {
        return this.isMinMaxDomainValid(min) && (min < this.max);
    }

    public checkMax(max: number): boolean {
        return this.isMinMaxDomainValid(max) && (this.min < max);
    }

    // includes checking min/max validity @TODO rename ?
    public checkStep(step: number): boolean {
        return this.isMinMaxValid && this.stepRefValue.intervalHasValue(step);
    }

    /**
     * Return true if single value is valid regarding the domain constraints
     */
    get isValueValid(): boolean {
        try {
            const v = this.paramValues.singleValue;
            this.checkValueAgainstDomain(v);
            return true;
        } catch (e) {
            return false;
        }
    }

    get isMinMaxValid(): boolean {
        return this.checkMinMax(this.min, this.max);
    }

    /**
     * Return true if current value is valid regarding the range constraints : min / max / step
     */
    public get isRangeValid(): boolean {
        switch (this._valueMode) {
            case ParamValueMode.LISTE:
                return this.isListValid;

            case ParamValueMode.MINMAX:
                // includes checking min/max
                return this.checkStep(this.step);
        }

        throw new Error(`ParamDefinition.isRangeValid() : valeur ${ParamValueMode[this._valueMode]}`
                      + `de ParamValueMode non prise en compte`);
    }

    /**
     * Root method to determine if a field value is valid, regarding the model constraints.
     *
     * Does not take care of input validation (ie: valid number, not empty...) because an
     * invalid input should never be set as a value; input validation is up to the GUI.
     *
     * In LINK mode :
     *  - if target is a parameter, checks validity of the target value against the local model constraints
     *  - if target is a result :
     *    - is result is already computed, checks validity of the target result against the local model constraints
     *    - if it is not, checks the "computability" of the target Nub (ie. validity of the Nub) to allow
     *      triggering chain computation
     */
    public get isValid(): boolean {
        switch (this._valueMode) {
            case ParamValueMode.SINGLE:
                return this.isValueValid;

            case ParamValueMode.MINMAX:
            case ParamValueMode.LISTE:
                return this.isRangeValid;

            case ParamValueMode.CALCUL:
                return true;

            case ParamValueMode.LINK:
                if (! this.isReferenceDefined()) {
                    return false;
                }
                // valuesIterator covers both target param and target result cases
                const iterator = this.valuesIterator;
                if (iterator) {
                    try {
                        for (const v of iterator) {
                            this.checkValueAgainstDomain(v);
                        }
                        return true;
                    } catch (e) {
                        return false;
                    }
                } else { // undefined iterator means target results are not computed yet
                    // check target Nub computability
                    return this.referencedValue.nub.isComputable();
                }
        }

        throw new Error(
            // tslint:disable-next-line:max-line-length
            `ParamDefinition.isValid() : valeur de ParamValueMode '${ParamValueMode[this._valueMode]}' non prise en charge`
        );
    }

    /**
     * Sets the current parameter to LINK mode, pointing to the given
     * symbol (might be a Parameter (computed or not) or an ExtraResult)
     * of the given Nub
     *
     * Prefer @see defineReferenceFromLinkedValue() whenever possible
     *
     * @param nub
     * @param symbol
     */
    public defineReference(nub: Nub, symbol: string) {
        /* console.log(`Defining reference from `
            + (this.parent ? `${this.nubUid}(${this.parentNub.constructor.name})` : "no_parent")
            + `/${this.symbol}`
            + ` to ${nub.uid}(${nub.constructor.name})/${symbol}`); */
        // prevent loops - beware that all Nubs must be registered in the
        // current session or this test will always fail
        if (! this.isAcceptableReference(nub, symbol)) {
            throw new Error("defineReference() : link target is not an available linkable value");
        }
        // clear current reference
        this.undefineReference();
        // find what the symbol points to
        if (nub) {
            // 1. extra result ?
            // - start here to avoid extra results being presented as
            // parameters by the iterator internally used by nub.getParameter()
            // - extra results with no family are not linkable
            if (Object.keys(nub.extraResultsFamilies).includes(symbol)) {
                this._referencedValue = new LinkedValue(nub, undefined, symbol);
            } else {
                // 2. is it a parameter (possibly in CALC mode) ?
                const p = nub.getParameter(symbol);
                if (p) {
                    this._referencedValue = new LinkedValue(nub, p, symbol);
                }
            }
        }
        if (this._referencedValue) {
            // set value mode
            this.valueMode = ParamValueMode.LINK;
        } else {
            throw new Error(`defineReference - could not find target for ${nub.uid}.${symbol}`);
        }
    }

    /**
     * Asks the Session for available linkabke values for the current parameter,
     * and returns true if given { nub / symbol } pair is part of them
     */
    public isAcceptableReference(nub: Nub, symbol: string) {
        const linkableValues = Session.getInstance().getLinkableValues(this);
        for (const lv of linkableValues) {
            if (lv.nub.uid === nub.uid && lv.symbol === symbol) {
                return true;
            }
        }
        return false;
    }

    public defineReferenceFromLinkedValue(target: LinkedValue) {
        this._referencedValue = target;
        this.valueMode = ParamValueMode.LINK;
    }

    public undefineReference() {
        this._referencedValue = undefined;
    }

    public isReferenceDefined(): boolean {
        return (this._referencedValue !== undefined);
    }

    /**
     * Returns an object representation of the Parameter's current state
     */
    public objectRepresentation(): { symbol: string, mode: string } {
        // parameter representation
        const paramRep: any = {
            symbol: this.symbol,
            mode: ParamValueMode[this._valueMode]
        };
        // adjust parameter representation depending on value mode
        switch (this._valueMode) {
            case ParamValueMode.SINGLE:
                paramRep.value = this.singleValue;
                break;

            case ParamValueMode.MINMAX:
                paramRep.min = this.min;
                paramRep.max = this.max;
                paramRep.step = this.step;
                break;

            case ParamValueMode.LISTE:
                paramRep.values = this.valueList;
                break;

            case ParamValueMode.LINK:
                paramRep.targetNub = this._referencedValue.nub.uid;
                paramRep.targetParam = this._referencedValue.symbol;
                break;
        }
        return paramRep;
    }

    /**
     * Fills the current Parameter, provided an object representation
     * @param obj object representation of a Parameter state
     */
    public loadObjectRepresentation(obj: any, setCalcMode: boolean = true): boolean {
        // set mode
        const mode: ParamValueMode = (ParamValueMode as any)[obj.mode]; // get enum index for string value

        // when loading parent Nub, setting mode to CALC would prevent ensuring
        // consistency when setting calculatedParam afterwards
        if (mode !== ParamValueMode.CALCUL || setCalcMode) {
            try {
                this.valueMode = mode;
            } catch (err) {
                console.error("loadObjectRepresentation: set valueMode error");
            }
        }

        // detect if this is the calculated param - used by Nub.loadObjectRepresentation()
        let isCalculated = false;

        // set value(s)
        switch (mode) {
            case ParamValueMode.SINGLE:
                this.singleValue = obj.value;
                break;

            case ParamValueMode.MINMAX:
                this.min = obj.min;
                this.max = obj.max;
                this.step = obj.step;
                break;

            case ParamValueMode.LISTE:
                this.valueList = obj.values;
                break;

            case ParamValueMode.CALCUL:
                // although calculated param is set at Nub level (see below),
                // it is detected as "the only parameter in CALC mode"
                isCalculated = true;
                break;

            case ParamValueMode.LINK:
                // formulaire dont le Nub est la cible du lien
                const destNub = Session.getInstance().findNubByUid(obj.targetNub);
                if (destNub) {
                    try {
                        this.defineReference(destNub, obj.targetParam);
                    } catch (err) {
                        console.error("loadObjectRepresentation: defineReference error");
                    }
                } // si la cible du lien n'existe pas, Session.fixLinks() est censé s'en occuper
                break;

            default:
                throw new Error(`session file : invalid value mode '${obj.valueMode}' in param object ${obj.symbol}`);
        }

        return isCalculated;
    }

    /**
     * Returns true if both the Nub UID and the symbol are identical
     */
    public equals(p: ParamDefinition): boolean {
        return (
            this.nubUid === p.nubUid
            && this.symbol === p.symbol
        );
    }

    /**
     * Returns true if this parameter's value can safely be linked to the
     * given parameter "src" (ie. without leading to circular dependencies)
     */
    public isLinkableTo(src: ParamDefinition): boolean {
        // did we loop back to the candidate parameter ?
        if (this.equals(src)) {
            // prevent infinite recursion
            return false;
        }

        if (this._valueMode === ParamValueMode.CALCUL) {
            // if my Nub doesn't depend on your result,
            // and my Nub doesn't already have a parameter linked to you,
            // you may depend on its result (me) !
            if (this._parent && this._parent.parent) {
                const myNub = (this.parentComputeNode as Nub);
                const ok = (
                    ! myNub.dependsOnNubResult(src.parentNub)
                    && ! myNub.dependsOnParameter(src)
                );
                return ok;
            }
        }

        if (this._valueMode === ParamValueMode.LINK && this.isReferenceDefined()) {
            // does the link point to an extra result ?
            if (this.referencedValue.isExtraResult()) {
                // if the target Nub doesn't depend on your result, you may depend on its extra result (me) !
                return (! this.referencedValue.nub.dependsOnNubResult(src.parentNub));

            } else { // the link points to a parameter, computed or not
                // recursively follow links, unless it loops back to the original parameter
                return (this.referencedValue.element as ParamDefinition).isLinkableTo(src);
            }
        }

        // SINGLE / MINMAX / LISTE parameters can always be linked without danger
        return true;
    }

    /**
     * Returns true if this parameter is ultimately (after followink links chain) linked
     * to a result or extra result of any of the given nubs.
     * If followLinksChain is false, will limit theexploration to the first target level only
     */
    public isLinkedToResultOfNubs(nubs: Nub[], followLinksChain: boolean = true): boolean {
        let linked = false;
        if (this._valueMode === ParamValueMode.LINK && this.isReferenceDefined()) {
            const targetNubUid = this.referencedValue.nub.uid;
            // is target a result ?
            if (this.referencedValue.isResult() || this.referencedValue.isExtraResult()) {
                outerloop:
                for (const y of nubs) {
                    if (y.uid === targetNubUid) {
                        // dependence found !
                        linked = true;
                        break outerloop;
                    }
                }
            } else { // target is a parameter
                // recursion ?
                if (followLinksChain) {
                    linked = (this.referencedValue.element as ParamDefinition).isLinkedToResultOfNubs(nubs);
                }
            }
        }
        return linked;
    }

    /**
     * Returns true if the current parameter is directly linked to the
     * Nub having UID "uid", its parent or any of its children
     * @param uid
     */
    public dependsOnNubFamily(uid: string): boolean {
        let linked = false;
        if (this._valueMode === ParamValueMode.LINK && this.isReferenceDefined()) {
            const ref = this._referencedValue;

            // direct reference ?
            if (ref.nub.uid === uid) {
                linked = true;

            // parent ?
            } else if (ref.nub.getParent() && ref.nub.getParent().uid === uid) {
                linked = true;

            // children ?
            } else if (ref.nub.getChildren().map(
                (c) => {
                    return c.uid;
                }).includes(uid)) {
                    linked = true;

            } // else no recursion, checking level 1 only
        }
        return linked;
    }

    // interface INamedIterableValues

    get name(): string {
        return this._symbol;
    }

    // interface IterableValues

    /**
     * Transparent proxy to own values iterator or targetted values iterator (in LINK mode)
     * @TODO rewrite more simply ?
     */
    public get valuesIterator(): INumberIterator {
        if (this.valueMode === ParamValueMode.LINK && this.isReferenceDefined()) {
            try {
                return this._referencedValue.getParamValues().valuesIterator;
            } catch (e) {
                // values are not computed yet
                return undefined;
            }
        } else {
            return this.paramValues.valuesIterator;
        }
    }

    /**
     * Returns true if there are more than 1 value associated to this parameter;
     * might be its own values (MINMAX / LISTE mode), or targetted values (LINK mode)
     */
    public get hasMultipleValues(): boolean {
        if (this._valueMode === ParamValueMode.LINK && this.isReferenceDefined()) {
            return this._referencedValue.hasMultipleValues();
        } else {
            if (this._valueMode === ParamValueMode.CALCUL) {
                return this.parentNub.resultHasMultipleValues();
            } else {
                return this.paramValues.hasMultipleValues;
            }
        }
    }

    public initValuesIterator(reverse: boolean = false): INumberIterator {
        return this.paramValues.getValuesIterator(reverse);
    }

    public get hasNext(): boolean {
        return this.paramValues.hasNext;
    }

    public next(): IteratorResult<number> {
        return this.paramValues.next();
    }

    public [Symbol.iterator](): IterableIterator<number> {
        return this.paramValues;
    }

    // interface IObservable

    /**
     * ajoute un observateur à la liste
     */
    public addObserver(o: Observer) {
        this._observable.addObserver(o);
    }

    /**
     * supprime un observateur de la liste
     */
    public removeObserver(o: Observer) {
        this._observable.removeObserver(o);
    }

    /**
     * notifie un événement aux observateurs
     */
    public notifyObservers(data: any, sender?: any) {
        this._observable.notifyObservers(data, sender);
    }

    /**
     * variable calculable par l'équation ?
     */
    public isAnalytical(): boolean {
        return this.calculability === ParamCalculability.EQUATION;
    }

    public clone(): ParamDefinition {
        const res = new ParamDefinition(this._parent, this._symbol,
            this.domain.clone(), this.singleValue, this._family, this.visible);
        res._calc = this._calc;
        return res;
    }

    /**
     * Asks parent Nub if any to ensure parameters consistency
     */
    protected ensureParametersConsistency(sourceParam: ParamDefinition,
                                          exceptMode: ParamValueMode|ParamValueMode[] = []
    ) {
        if (this.parentComputeNode && this.parentComputeNode instanceof Nub) {
            this.parentComputeNode.ensureParametersConsistency(sourceParam, exceptMode);
        }
    }

    /**
     * notification envoyée après la modification de la valeur du paramètre
     */
    private notifyValueModified(sender: any) {
        this.notifyObservers(
            { action: "paramdefinitionAfterValue" }, sender
        );
    }

    /**
     * vérifie si un min/max est valide par rapport au domaine de définition
     */
    private isMinMaxDomainValid(v: number): boolean {
        if (v === undefined) {
            return false;
        }
        if (this._valueMode === ParamValueMode.MINMAX) {
            try {
                this.checkValueAgainstDomain(v);
            } catch (e) {
                return false;
            }
        }
        return true;
    }

    private checkMinMax(min: number, max: number): boolean {
        return this.isMinMaxDomainValid(min) && this.isMinMaxDomainValid(max) && (min < max);
    }

    private get isListValid(): boolean {
        if (this.paramValues.valueList === undefined) {
            return false;
        }

        for (const v of this.paramValues.valueList) {
            try {
                this.checkValueAgainstDomain(v);
            } catch (e) {
                return false;
            }
        }
        return true;
    }

    /**
     * Throws an error if this.paramValues._valueMode is not the expected value mode.
     * In LINK mode, the tested valueMode is the mode of the ultimately targetted
     * set of values (may never be LINK)
     */
    private checkValueMode(expected: ParamValueMode[]) {
        if (! expected.includes(this.valueMode)) {
            throw new Error(
                `ParamDefinition : mode de valeurs ${ParamValueMode[this.valueMode]} incorrect`
            );
        }
    }

}
