import { INamedObject, IObjectWithFamily } from "../jalhyd_object";
import { ArrayReverseIterator } from "../util/iterator";
import { ParamValueMode } from "./param-value-mode";
import { ParamValues } from "./param-values";

/**
 * itérateur sur des nombres
 */
export interface INumberIterator extends IterableIterator<number> {
    /**
     * @return true si il reste des valeurs à parcourir
     */
    readonly hasNext: boolean;

    /**
     * valeur courante cad résultat du précédent appel à next().
     * contrairement à next(), on peut l'appeler plusieurs fois sans modifier l'état de l'itérateur
     */
    readonly currentValue: number;

    /**
     * prochaine valeur
     */
    next(): IteratorResult<number>;

    count(): number;
}

/**
 * interface implémentée par les objets pouvant posséder/renvoyer un itérateur sur une série de valeurs numériques
 */
export interface IterableValues extends INumberIterator {
    /**
     * crée un nouvel itérateur sur les valeurs
     */
    readonly valuesIterator: INumberIterator;

    /**
     * true si la série de valeurs a plus d'une valeur
     */
    readonly hasMultipleValues: boolean;

    /**
     * initialise un itérateur interne sur les valeurs implicitement utilisé par hasNext() et next()
     */
    initValuesIterator(reverse: boolean): INumberIterator;
}

/**
 * objets pouvant avoir un nom et un itérateur sur une série de valeurs numériques
 */
export interface INamedIterableValues extends INamedObject, IObjectWithFamily, IterableValues {
}

/**
 * itérateur sur les (ou la) valeurs prises par un ParamValues
 */
export class ParamValueIterator implements INumberIterator {
    /**
     * paramètre à itérer
     */
    private _param: ParamValues;

    /**
     * true si les valeurs sont fournies de max à min (ParamValueMode.MINMAX)
     */
    private _reverse: boolean;

    /**
     * valeur courante
     */
    private _current: number;

    private _index: number;

    constructor(prm: ParamValues, reverse: boolean = false) {
        prm.check();
        this._param = prm;
        this.reset(reverse);
    }

    /**
     * Returns the number of elements in the iterator
     */
    public count() {
        let length = 0;
        for (const i of this) {
            length++;
        }
        this.reset(false);
        return length;
    }

    public reset(reverse: boolean) {
        this._reverse = reverse;

        switch (this._param.valueMode) {
            case ParamValueMode.SINGLE:
                this._index = 0;
                break;

            case ParamValueMode.MINMAX:
                if (reverse) {
                    this._index = this._param.max;
                } else {
                    this._index = this._param.min;
                }
                break;

            case ParamValueMode.LISTE:
                this._index = 0;
                break;

            default:
                // tslint:disable-next-line:max-line-length
                throw new Error(`ParamValueIterator : mode de génération de valeurs ${ParamValueMode[this._param.valueMode]} incorrect`);
        }
    }

    public get hasNext(): boolean {
        switch (this._param.valueMode) {

            case ParamValueMode.SINGLE:
                return this._index === 0;

            case ParamValueMode.MINMAX:
                const end = this._reverse ?
                    this._index < this._param.min - this._param.step * 1E-7 :
                    this._index > this._param.max + this._param.step * 1E-7;
                return !end;

            case ParamValueMode.LISTE:
                return this._index < this._param.valueList.length;

            default:
                throw new Error(`ParamValueIterator.hasNext() : erreur interne`);
        }
    }

    public next(): IteratorResult<number> {
        switch (this._param.valueMode) {

            case ParamValueMode.SINGLE:
                if (this.hasNext) {
                    this._current = this._param.singleValue;
                    this._index++;
                    return {
                        done: false,
                        value: this._current
                    };
                } else {
                    return {
                        done: true,
                        value: undefined
                    };
                }

            case ParamValueMode.MINMAX:
                this._current = this._index;
                if (this.hasNext) {
                    if (this._reverse) {
                        this._index -= this._param.step;
                    } else {
                        this._index += this._param.step;
                    }
                    return {
                        done: false,
                        value: this._current
                    };
                } else {
                    return {
                        done: true,
                        value: undefined
                    };
                }

            case ParamValueMode.LISTE:
                if (this.hasNext) {
                    this._current = this._param.valueList[this._index++];
                    return {
                        done: false,
                        value: this._current
                    };
                } else {
                    return {
                        done: true,
                        value: undefined
                    };
                }

            default:
                throw new Error(`ParamValueIterator.next() : erreur interne`);
        }
    }

    public get currentValue(): number {
        return this._current;
    }

    // interface IterableIterator

    public [Symbol.iterator](): IterableIterator<number> {
        return this;
    }
}

/**
 * itérateur sur les valeurs prises par un tableau
 */
// tslint:disable-next-line:max-classes-per-file
export class NumberArrayIterator implements INumberIterator {
    private _it: IterableIterator<number>;

    private _index: number;

    /**
     * valeur courante
     */
    private _current: number;

    constructor(private _arr: number[]) {
        this._it = this._arr[Symbol.iterator](); // crée un itérateur à partir d'un tableau
        this._index = 0;
    }

    public count() {
        return this._arr.length;
    }

    public get hasNext(): boolean {
        return this._index < this._arr.length;
    }

    public next(): IteratorResult<number> {
        this._index++;
        const res = this._it.next();
        if (!res.done) {
            this._current = res.value;
        }
        return res;
    }

    public get currentValue(): number {
        return this._current;
    }

    // interface IterableIterator

    public [Symbol.iterator](): IterableIterator<number> {
        return this;
    }
}

/**
 * itérateur sur les valeurs prises par un tableau (parcourues depuis la fin)
 */
// tslint:disable-next-line:max-classes-per-file
export class NumberArrayReverseIterator extends ArrayReverseIterator<number> implements INumberIterator {
    private _count: number;

    /**
     * valeur courante
     */
    private _current: number;

    constructor(arr: number[]) {
        super(arr);
        this._count = 0;
    }

    public count() {
        return this._arr.length;
    }

    public get hasNext(): boolean {
        return this._count < super._arr.length;
    }

    public next(): IteratorResult<number> {
        this._count++;
        const res = super.next();
        if (!res.done) {
            this._current = res.value;
        }
        return res;
    }

    public get currentValue(): number {
        return this._current;
    }

    // interface IterableIterator

    public [Symbol.iterator](): IterableIterator<number> {
        return this;
    }
}
