import { INumberIterator, IterableValues, ParamValueIterator } from "./param-value-iterator";
import { ParamValueMode } from "./param-value-mode";

/**
 * Represents the value(s) taken by a Parameter, along with and depending on
 * the value mode; linked values are not managed here, only the LINK value
 * mode is defined here
 */
export class ParamValues implements IterableValues {

    /**
     * usually set by enclosing ParamDefinition
     */
    public valueMode: ParamValueMode;

    /** valeur courante (éventuellement non définie) indépendemment du mode */
    public currentValue: number;

    /** valeur min dans le cas ParamValueMode.MINMAX */
    public min: number;

    /** valeur max dans le cas ParamValueMode.MINMAX */
    public max: number;

    /** pas de progression dans le cas ParamValueMode.MINMAX */
    public step: number;

    /** liste de valeurs dans le cas ParamValueMode.LISTE */
    public valueList: number[];

    /** valeur numérique (éventuellement non définie) dans le mode "valeur unique" (ParamValueMode.SINGLE) */
    private _singleValue: number;

    /**
     * itérateur courant
     */
    private _iterator: INumberIterator;

    public get singleValue(): number {
        return this._singleValue;
    }

    public set singleValue(v: number) {
        this._singleValue = v;
        this.currentValue = v;
    }

    public count() {
        return this._iterator.count();
    }

    public setValues(o: number | any, max?: number, step?: number) {
        if (typeof (o) === "number") {
            if (max === undefined) {
                this.valueMode = ParamValueMode.SINGLE;
                this.singleValue = o;
            } else {
                this.valueMode = ParamValueMode.MINMAX;
                this.min = o;
                this.max = max;
                this.step = step;
            }
        } else if (Array.isArray(o)) {
            this.valueMode = ParamValueMode.LISTE;
            this.valueList = o;
        } else {
            throw new Error(`ParamValues.setValues() :  appel invalide`);
        }
    }

    public check() {
        switch (this.valueMode) {
            case ParamValueMode.SINGLE:
                if (this.singleValue === undefined) {
                    throw new Error(`ParamValues : valeur fixe non définie`);
                }
                break;

            case ParamValueMode.MINMAX:
                if (this.min === undefined) {
                    throw new Error(`ParamValues : valeur min non définie`);
                }
                if (this.max === undefined) {
                    throw new Error(`ParamValues : valeur max non définie`);
                }
                if (this.step === undefined) {
                    throw new Error(`ParamValues : valeur du pas non définie`);
                }
                if (this.min > this.max) {
                    throw new Error(`ParamValues : min > max`);
                }
                break;

            case ParamValueMode.LISTE:
                if (this.valueList === undefined) {
                    throw new Error(`ParamValues : liste de valeurs non définie`);
                }
                break;

            case ParamValueMode.LINK:
            case ParamValueMode.CALCUL:
            default:
                break;
        }
    }

    // -- iterator-related methods

    /**
     * crée un ParamValueIterator
     * @param reverse true si on veut itérer max->min ou depuis la fin de la liste
     */
    public getValuesIterator(reverse: boolean = false): INumberIterator {
        return new ParamValueIterator(this, reverse);
    }

    // interface IterableValues

    public get valuesIterator(): INumberIterator {
        return this.getValuesIterator();
    }

    public get hasMultipleValues(): boolean {
        const it = this.getValuesIterator();
        if (it) {
            let n = 0;
            for (const v of it) {
                n++;
                if (n > 1) {
                    break;
                }
            }
            return n > 1;
        } else {
            return false;
        }
    }

    public initValuesIterator(reverse: boolean = false): INumberIterator {
        switch (this.valueMode) {
            case ParamValueMode.LISTE:
            case ParamValueMode.MINMAX:
                this._iterator = this.getValuesIterator(reverse);
                break;

            default:
                throw new Error(`ParamValues : mode de valeurs ${ParamValueMode[this.valueMode]} incorrect`);
        }

        return this._iterator;
    }

    /**
     * @return true si il reste des valeurs à parcourir par l'itérateur courant
     */
    public get hasNext(): boolean {
        return this._iterator.hasNext;
    }

    /**
     * fixe la valeur courante à la prochaine valeur à parcourir par l'itérateur courant
     * @return prochaine valeur à parcourir par l'itérateur courant
     */
    public next(): IteratorResult<number> {
        let res;
        res = this._iterator.next();
        if (!res.done) {
            this.currentValue = res.value;
        }
        return res;
    }

    public [Symbol.iterator](): IterableIterator<number> {
        return this;
    }
}
