import { Message, MessageCode } from "../util/message";
import { Result } from "../util/result";
import { acNewton } from "./newton";
import { acSection, ParamsSection } from "./section_type";

/**
 * Calcul de la hauteur critique
 */
// tslint:disable-next-line:class-name
export class cHautCritique extends acNewton {
        /**
         * Section sur laquuelle porte le calcul
         */
        // tslint:disable-next-line:variable-name
        private Sn: acSection;

        /**
         * Constructeur de la classe
         * @param Sn Section sur laquelle on fait le calcul
         */
        // tslint:disable-next-line:variable-name
        constructor(Sn: acSection, maxIter: number, dbg: boolean = false) {
                super(Sn.prms, maxIter, dbg);
                this.Sn = Sn.clone();
        }

        /**
         * Calcul de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        public CalcFn(rX: number): Result {
                const rS: Result = this.Sn.CalcSection("S", rX);
                if (!rS.ok) {
                        return rS;
                }

                // Calcul de la fonction
                // if (this.Sn.CalcSection("S", rX) !== 0)
                if (rS.vCalc === 0) {
                        return new Result(new Message(MessageCode.ERROR_SECTION_SURFACE_NULLE));
                }

                // tslint:disable-next-line:max-line-length
                // return (Math.pow(this.Sn.prms.Q.v, 2) * this.Sn.CalcSection("B", rX) / Math.pow(this.Sn.CalcSection("S", rX), 3) / ParamsSection.G - 1);
                const rB: Result = this.Sn.CalcSection("B", rX);
                if (!rB.ok) {
                        return rB;
                }

                const rS2: Result = this.Sn.CalcSection("S", rX);
                if (!rS2.ok) {
                        return rS2;
                }

                const v = (Math.pow(this.Sn.prms.Q.v, 2) * rB.vCalc / Math.pow(rS2.vCalc, 3) / ParamsSection.G - 1);
                return new Result(v);

                // return Infinity;
        }

        /**
         * Calcul analytique de la dérivée de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        public CalcDer(rX: number): Result {
                // let S = this.Sn.CalcSection("S");
                const rS: Result = this.Sn.CalcSection("S");
                if (!rS.ok) {
                        return rS;
                }
                const S = rS.vCalc;
                // if (S !== 0) {
                if (S === 0) {
                        return new Result(new Message(MessageCode.ERROR_SECTION_SURFACE_NULLE));
                }

                // let B = this.Sn.CalcSection("B");
                const rB: Result = this.Sn.CalcSection("B");
                if (!rB.ok) {
                        return rB;
                }
                const B = rB.vCalc;

                const rDB: Result = this.Sn.CalcSection("dB");
                if (!rDB.ok) {
                        return rDB;
                }

                // L'initialisation à partir de rX a été faite lors de l'appel à CalcFn
                // let Der = (this.Sn.CalcSection("dB") * S - 3 * B * B);
                // tslint:disable-next-line:variable-name
                const Der = (rDB.vCalc * S - 3 * B * B);
                const v = Math.pow(this.Sn.prms.Q.v, 2) / ParamsSection.G * Der / Math.pow(S, 4);
                return new Result(v);
                // }

                // return Infinity;
        }
}

/**
 * Calcul de la hauteur normale
 */
// tslint:disable-next-line:max-classes-per-file class-name
export class cHautNormale extends acNewton {
        /**
         * Section sur laquuelle porte le calcul
         */
        // tslint:disable-next-line:variable-name
        private Sn: acSection;

        /**
         * Débit connu
         */
        private Q: number;

        /**
         * Coefficient de Strickler
         */
        // tslint:disable-next-line:variable-name
        private Ks: number;

        /**
         * Pente du fond
         */
        // tslint:disable-next-line:variable-name
        private If: number;

        /**
         * Constructeur de la classe
         * @param oSn Section sur laquelle on fait le calcul
         */
        // tslint:disable-next-line:variable-name
        constructor(Sn: acSection, maxIter: number, dbg: boolean = false) {
                super(Sn.prms, maxIter, dbg);
                this.Sn = Sn;
                this.Q = Sn.prms.Q.v;
                this.Ks = Sn.prms.Ks.v;
                this.If = Sn.prms.If.v;
        }

        /**
         * Calcul de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        public CalcFn(rX: number): Result {
                // Calcul de la fonction
                // tslint:disable-next-line:max-line-length
                // return (this.Q - this.Ks * Math.pow(this.Sn.CalcSection("R", rX), 2 / 3) * this.Sn.CalcSection("S", rX) * Math.sqrt(this.If));
                const rR: Result = this.Sn.CalcSection("R", rX);
                if (!rR.ok) {
                        return rR;
                }

                const rS: Result = this.Sn.CalcSection("S", rX);
                if (!rS.ok) {
                        return rS;
                }

                const v = (this.Q - this.Ks * Math.pow(rR.vCalc, 2 / 3) * rS.vCalc * Math.sqrt(this.If));
                return new Result(v);
        }

        /**
         * Calcul analytique de la dérivée de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        public CalcDer(rX: number): Result {
                const rDR: Result = this.Sn.CalcSection("dR");
                if (!rDR.ok) {
                        return rDR;
                }

                const rR: Result = this.Sn.CalcSection("R");
                if (!rR.ok) {
                        return rR;
                }

                const rS: Result = this.Sn.CalcSection("S");
                if (!rS.ok) {
                        return rS;
                }

                // L'initialisation a été faite lors de l'appel à CalcFn

                // let Der = 2 / 3 * this.Sn.CalcSection("dR") * Math.pow(this.Sn.CalcSection("R"), -1 / 3) * this.Sn.CalcSection("S");
                // tslint:disable-next-line:variable-name
                let Der = 2 / 3 * rDR.vCalc * Math.pow(rR.vCalc, -1 / 3) * rS.vCalc;

                const rR2: Result = this.Sn.CalcSection("R");
                if (!rR2.ok) {
                        return rR2;
                }

                const rB: Result = this.Sn.CalcSection("B");
                if (!rB.ok) {
                        return rB;
                }

                // Der = Der + Math.pow(this.Sn.CalcSection("R"), 2 / 3) * this.Sn.CalcSection("B");
                Der = Der + Math.pow(rR2.vCalc, 2 / 3) * rB.vCalc;

                Der = Der * -this.Ks * Math.sqrt(this.If);
                return new Result(Der);
        }
}

/**
 * Calcul de la hauteur correspondante (charge égale)
 */
// tslint:disable-next-line:max-classes-per-file class-name
export class cHautCorrespondante extends acNewton {
        /**
         * Tirant d'eau connu
         */
        private Y: number;

        /**
         * 1/S^2 associé au tirant d'eau connu
         */
        private _rS2: Result;

        /**
         * Section contenant les données de la section avec la hauteur à calculer
         */
        // tslint:disable-next-line:variable-name
        private Sn: acSection;

        /**
         * Constante de gravité
         */
        private rQ2G: number;

        /**
         * Constructeur de la classe
         * @param oSn Section sur laquelle on fait le calcul
         */
        // tslint:disable-next-line:variable-name
        constructor(Sn: acSection, maxIter: number, dbg: boolean = false) {
                super(Sn.prms, maxIter, dbg);
                this.Y = Sn.prms.Y.v;
                //                this.rS2 = Math.pow(Sn.CalcSection("S"), -2);
                this.Sn = Sn;
                // tslint:disable-next-line:no-unused-expression
                this.rS2;  // pour initialiser la valeur @WTF (utilise le getter)
                this.rQ2G = Math.pow(Sn.prms.Q.v, 2) / (2 * ParamsSection.G);
        }

        private get rS2(): Result {
                if (this._rS2 === undefined) {
                        const rS = this.Sn.CalcSection("S");
                        if (rS.ok) {
                                const v = Math.pow(rS.vCalc, -2);
                                this._rS2 = new Result(v);
                        } else {
                                this._rS2 = rS;
                        }
                }
                return this._rS2;
        }

        /**
         * Calcul de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        public CalcFn(rX: number): Result {
                if (!this.rS2.ok) {
                        return this.rS2;
                }

                const rS: Result = this.Sn.CalcSection("S", rX);
                if (!rS.ok) {
                        return rS;
                }

                // Calcul de la fonction
                // return this.Y - rX + (this.rS2 - Math.pow(this.Sn.CalcSection("S", rX), -2)) * this.rQ2G;
                const v = this.Y - rX + (this.rS2.vCalc - Math.pow(rS.vCalc, -2)) * this.rQ2G;
                return new Result(v);
        }

        /**
         * Calcul analytique de la dérivée de la fonction dont on protected function cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        public CalcDer(rX: number): Result {
                // let S = this.Sn.CalcSection("S");
                const rS: Result = this.Sn.CalcSection("S");
                if (!rS.ok) {
                        return rS;
                }
                const S = rS.vCalc;

                // L'initialisation a été faite lors de l'appel à CalcFn
                // if (S !== 0)
                if (S === 0) {
                        return new Result(new Message(MessageCode.ERROR_SECTION_SURFACE_NULLE));
                }

                const rB: Result = this.Sn.CalcSection("B");
                if (!rB.ok) {
                        return rB;
                }

                // return -1 + 2 * this.rQ2G * this.Sn.CalcSection("B") / Math.pow(S, 3);
                const v = -1 + 2 * this.rQ2G * rB.vCalc / Math.pow(S, 3);
                return new Result(v);

                // return Infinity;
        }
}

/**
 * Calcul de la hauteur conjuguée (Impulsion égale)
 */
// tslint:disable-next-line:max-classes-per-file class-name
export class cHautConjuguee extends acNewton {
        /** Section contenant les données de la section avec la hauteur à calculer */
        // tslint:disable-next-line:variable-name
        private Sn: acSection;

        /** Carré du débit */
        private rQ2: number;

        /** Surface hydraulique associée au tirant d'eau connu */
        private rS: Result;

        /** SYg associée au tirant d'eau connu */
        private rSYg: Result;

        /**
         * Constructeur de la classe
         * @param oSn Section sur laquelle on fait le calcul
         */
        constructor(oSn: acSection, maxIter: number, dbg: boolean = false) {
                super(oSn.prms, maxIter, dbg);
                this.rQ2 = Math.pow(oSn.prms.Q.v, 2);
                this.Sn = oSn;
                this.rS = oSn.CalcSection("S");
                this.rSYg = oSn.CalcSection("SYg");
        }

        /**
         * Calcul de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        public CalcFn(rX: number) {
                if (!this.rS.ok) {
                        return this.rS;
                }

                if (!this.rSYg.ok) {
                        return this.rSYg;
                }

                const rS2: Result = this.Sn.CalcSection("S", rX);
                if (!rS2.ok) {
                        return rS2;
                }

                // Réinitialisation des paramètres hydrauliques de this.Sn avec l'appel this.Sn.CalcSection("S',rX)
                // if (this.rS > 0 && this.Sn.CalcSection("S", rX) > 0) {
                if (this.rS.vCalc > 0 && rS2.vCalc > 0) {
                        // let Fn = this.rQ2 * (1 / this.rS - 1 / this.Sn.CalcSection("S"));
                        const rS3: Result = this.Sn.CalcSection("S");
                        if (!rS3.ok) {
                                return rS3;
                        }

                        // tslint:disable-next-line:variable-name
                        const Fn = this.rQ2 * (1 / this.rS.vCalc - 1 / rS3.vCalc);

                        const rSYg = this.Sn.CalcSection("SYg");
                        if (!rSYg.ok) {
                                return rSYg;
                        }

                        // return Fn + ParamsSection.G * (this.rSYg - this.Sn.CalcSection("SYg"));
                        const v = Fn + ParamsSection.G * (this.rSYg.vCalc - rSYg.vCalc);
                        return new Result(v);
                }

                return new Result(new Message(MessageCode.ERROR_SECTION_PERIMETRE_NUL));
                // return -Infinity;
        }

        /**
         * Calcul analytique de la dérivée de la fonction dont on cherche le zéro
         * @param rX Variable dont dépend la fonction
         */
        public CalcDer(rX: number): Result {
                if (!this.rS.ok) {
                        return this.rS;
                }

                //                let S = this.Sn.CalcSection("S");
                const rS2: Result = this.Sn.CalcSection("S");
                if (!rS2.ok) {
                        return rS2;
                }
                const S = rS2.vCalc;

                // L'initialisation a été faite lors de l'appel à CalcFn
                // if (this.rS > 0 && S > 0) {
                if (this.rS.vCalc > 0 && S > 0) {
                        const rDS: Result = this.Sn.CalcSection("dS");
                        if (!rDS.ok) {
                                return rDS;
                        }

                        // let Der = this.rQ2 * this.Sn.CalcSection("dS") * Math.pow(S, -2);
                        // tslint:disable-next-line:variable-name
                        const Der = this.rQ2 * rDS.vCalc * Math.pow(S, -2);

                        const rDYG: Result = this.Sn.CalcSection("dSYg", rX);
                        if (!rDYG.ok) {
                                return rDYG;
                        }

                        // return Der - ParamsSection.G * this.Sn.CalcSection("dSYg", rX);
                        const v = Der - ParamsSection.G * rDYG.vCalc;
                        return new Result(v);
                }

                return new Result(new Message(MessageCode.ERROR_SECTION_PERIMETRE_NUL));
                // return -Infinity;
        }
}
