import { ComputeNodeType } from "../compute-node";
import { ParamCalculability, ParamDefinition, ParamFamily } from "../param/param-definition";
import { ParamDomainValue } from "../param/param-domain";
import { Result } from "../util/result";
import { acSection, ParamsSection } from "./section_type";

export class ParamsSectionTrapez extends ParamsSection {
        private _LargeurFond: ParamDefinition; // Largeur au fond
        private _Fruit: ParamDefinition; // Fruit des berges

        constructor(rLargeurFond: number, rFruit: number, rY: number, rKs: number,
                    rQ: number, rIf: number, rYB: number) {

                super(rY, undefined, rKs, rQ, rIf, rYB);
                this._LargeurFond = new ParamDefinition(this, "LargeurFond", ParamDomainValue.POS_NULL,
                        rLargeurFond, ParamFamily.WIDTHS);
                this._Fruit = new ParamDefinition(this, "Fruit", ParamDomainValue.POS_NULL, rFruit);

                this.addParamDefinition(this._LargeurFond);
                this.addParamDefinition(this._Fruit);
                // hide params
                this.LargeurBerge.visible = false;
                this.LargeurBerge.undefineFamily(); // or else something might get linked to it although it is undefined
        }

        /**
         * Largeur au fond
         */
        get LargeurFond(): ParamDefinition {
                return this._LargeurFond;
        }

        /**
         * Fruit des berges
         */
        get Fruit(): ParamDefinition {
                return this._Fruit;
        }
}

/**
 * Calculs de la section trapézoïdale
 */
// tslint:disable-next-line:max-classes-per-file class-name
export class cSnTrapez extends acSection {

        get prms(): ParamsSectionTrapez {
                return this._prms as ParamsSectionTrapez;
        }
        constructor(prms: ParamsSectionTrapez, dbg: boolean = false) {
                super(prms, dbg);
                this._nodeType = ComputeNodeType.SectionTrapeze;
        }

        protected setParametersCalculability() {
                super.setParametersCalculability();
                this.prms.LargeurFond.calculability = ParamCalculability.DICHO;
                this.prms.Fruit.calculability = ParamCalculability.DICHO;
        }

        protected Calc_B(): Result {
                let v;
                if (this.isDebordement()) {
                        v = this.prms.LargeurFond.v + 2 * this.prms.Fruit.v * this.prms.YB.v;
                        return new Result(v);
                }

                v = this.prms.LargeurFond.v + 2 * this.prms.Fruit.v * this.prms.Y.v;
                return new Result(v);
        }

        /**
         * Calcul du périmètre mouillé
         * @return Périmètre mouillé (m)
         */
        protected Calc_P(): Result {
                let v;
                if (this.isDebordement()) {
                        // return this.CalcGeo("P") + super.Calc_P_Debordement(this.prms.Y.v - this.prms.YB.v);

                        const rGeoP: Result = this.CalcGeo("P");
                        if (!rGeoP.ok) {
                                return rGeoP;
                        }

                        const rPDeb: Result = super.Calc_P_Debordement(this.prms.Y.v - this.prms.YB.v);
                        if (!rPDeb.ok) {
                                return rPDeb;
                        }

                        v = rGeoP.vCalc + rPDeb.vCalc;
                        return new Result(v);
                }

                v = this.prms.LargeurFond.v + 2 * Math.sqrt(1 + Math.pow(this.prms.Fruit.v, 2)) * this.prms.Y.v;
                return new Result(v);
        }

        /**
         * Calcul de la surface mouillée
         * @return Surface mouillée (m2)
         */
        protected Calc_S(): Result {
                let v;
                if (this.isDebordement()) {
                        // return this.CalcGeo("S") + super.Calc_S_Debordement(this.prms.Y.v - this.prms.YB.v);
                        const rGeoS: Result = this.CalcGeo("S");
                        if (!rGeoS.ok) {
                                return rGeoS;
                        }

                        const rSDeb: Result = super.Calc_S_Debordement(this.prms.Y.v - this.prms.YB.v);
                        if (!rSDeb.ok) {
                                return rSDeb;
                        }

                        v = rGeoS.vCalc + rSDeb.vCalc;
                        return new Result(v);
                }

                v = this.prms.Y.v * (this.prms.LargeurFond.v + this.prms.Fruit.v * this.prms.Y.v);
                return new Result(v);
        }

        /**
         * Calcul de dérivée de la surface hydraulique par rapport au tirant d'eau.
         * @return dS
         */
        protected Calc_dS(): Result {
                if (this.isDebordement()) {
                        return super.Calc_dS();
                }

                const v = this.prms.LargeurFond.v + 2 * this.prms.Fruit.v * this.prms.Y.v;
                return new Result(v);
        }

        /**
         * Calcul de dérivée du périmètre hydraulique par rapport au tirant d'eau.
         * @return dP
         */
        protected Calc_dP(): Result {
                if (this.isDebordement()) {
                        return super.Calc_dP_Debordement();
                }

                const v = 2 * Math.sqrt(1 + this.prms.Fruit.v * this.prms.Fruit.v);
                return new Result(v);
        }

        /**
         * Calcul de dérivée de la largeur au miroir par rapport au tirant d'eau.
         * @return dB
         */
        protected Calc_dB(): Result {
                if (this.isDebordement()) {
                        return super.Calc_dB_Debordement();
                }

                const v = 2 * this.prms.LargeurFond.v * this.prms.Fruit.v;
                return new Result(v);
        }

        /**
         * Calcul de la distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_SYg(): Result {
                const v = (this.prms.LargeurFond.v / 2 + this.prms.Fruit.v * this.prms.Y.v / 3)
                        * Math.pow(this.prms.Y.v, 2);
                return new Result(v);
        }

        /**
         * Calcul de la dérivée de la distance du centre de gravité de la section à la surface libre
         * multiplié par la surface hydraulique
         * @return S x Yg
         */
        protected Calc_dSYg(): Result {
                // tslint:disable-next-line:variable-name
                let SYg = this.prms.Fruit.v / 3 * Math.pow(this.prms.Y.v, 2);
                SYg += (this.prms.LargeurFond.v / 2 + this.prms.Fruit.v * this.prms.Y.v / 3) * 2 * this.prms.Y.v;
                return new Result(SYg);
        }

        private isDebordement(): boolean {
                return this.prms.Y.v > this.prms.YB.v;
        }
}
