import { CalculatorType } from "../compute-node";
import { ParamCalculability, ParamFamily } from "../param/param-definition";
import { Result } from "../util/result";
import { DeverParams } from "./dever_params";
import { ParallelStructure } from "./parallel_structure";
import { loiAdmissiblesDever, LoiDebit } from "./structure_props";

export { DeverParams };

export class Dever extends ParallelStructure {
    constructor(prms: DeverParams, dbg: boolean = false) {
        super(prms, dbg);
        this._calcType = CalculatorType.Dever;
    }

    /**
     * paramètres castés au bon type
     */
    get prms(): DeverParams {
        return this._prms as DeverParams;
    }

    public getLoisAdmissibles(): { [key: string]: LoiDebit[]; } {
        return loiAdmissiblesDever;
    }

    /**
     * Calcul pour tous les paramètres
     * @param sVarCalc Paramètre à calculer
     * @returns Résultat avec résultats complémentaires et message
     */
    public Equation(sVarCalc: string): Result {
        const r: Result = super.Equation(sVarCalc);
        let QT: number;
        if (sVarCalc === "Q") {
            QT = r.vCalc;
        } else {
            QT = this.prms.Q.v;
        }
        // Vitesse dans le cours d'eau amont
        r.extraResults.V = QT / (this.prms.BR.v * Math.max(0, this.prms.Z1.v - this.prms.ZR.v));
        // Energie cinétique dans le cours d'eau amont
        r.extraResults.Ec = r.extraResults.V * r.extraResults.V / (2 * 9.81);
        // Calcul du rapport des aires sur les seuils et de l'aire du cours d'eau amont
        const rA: number = this.calcA() / (this.prms.BR.v * Math.max(0, this.prms.Z1.v - this.prms.ZR.v));
        // Calcul de Cv
        r.extraResults.Cv = this.calcCv(rA);
        // Calcul de CV.QT
        r.extraResults.CvQT = r.extraResults.Cv * QT;
        return r;
    }

    /**
     * paramétrage de la calculabilité des paramètres
     */
    protected setParametersCalculability() {
        super.setParametersCalculability();
        this.prms.BR.calculability = ParamCalculability.FREE;
        this.prms.ZR.calculability = ParamCalculability.FREE;
        // On supprime Z2 de l'interface car régime dénoyé uniquement
        this.prms.Z2.calculability = ParamCalculability.NONE;
    }

    protected setExtraResultsFamilies() {
        this._extraResultsFamilies = {
            CvQT: ParamFamily.FLOWS
        };
    }

    /**
     * Calcul de la somme des aires des ouvrages.
     * @note N.B. : doit être appelé après calcul si le calcul porte sur Z1 ou
     */
    private calcA(): number {
        let A: number = 0;
        for (const st of this.structures) {
            A += st.calcA();
        }
        return A;
    }

    /**
     * Calcul du facteur de la correction du coefficient de débit dû à la vitesse d'approche
     * @param rA Rapport entre l'aire d'écoulement des seuils et l'aire de la section amont
     */
    private calcCv(rA: number): number {
        const aA: number[] = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8];
        const aCv: number[] = [1, 1.005, 1.015, 1.022, 1.035, 1.065, 1.09, 1.14, 1.21];
        // tslint:disable-next-line:prefer-for-of
        for (let i = 1; i < aA.length; i++) {
            if (rA < aA[i]) {
                const k: number = (rA - aA[i - 1]) / (aA[i] - aA[i - 1]);
                return aCv[i - 1] * (1 - k) + k * aCv[i];
            }
        }
        // Rapport des aires supérieur à 0.8 => Cv = 1.21
        return aCv[aCv.length - 1];
    }

}
