import { ParallelStructure } from "./parallel_structure";

export enum StructureType {
    SeuilRectangulaire,
    VanneRectangulaire,
    Orifice,
    SeuilTriangulaire,
    SeuilTriangulaireTrunc
    // VanneCirculaire,
    // VanneTrapezoidale, SeuilTrapezoidal
}

export enum LoiDebit {
    // loi de débit Déversoir / Orifice Cemagref 1988 (pour Vanne Rectangulaire)
    GateCem88d,
    // loi de débit Déversoir / Vanne de fond Cemagref 1988 (pour Vanne Rectangulaire)
    GateCem88v,
    // loi de débit Déversoir / Orifice Cemagref 1988 (pour Seuil Rectangulaire)
    WeirCem88d,
    // loi de débit Déversoir / Vanne de fond Cemagref 1988 (pour Seuil Rectangulaire)
    WeirCem88v,
    // loi de débit Cunge 1980
    Cunge80,
    // loi de débit pour vanne dénoyée
    RectangularOrificeFree,
    // loi de débit pour vanne noyée
    RectangularOrificeSubmerged,
    // loi de débit pour seuil dénoyé
    WeirFree,
    // Loi Kindsvater-Carter et Villemonte
    KIVI,
    // Loi de débit seuil triangulaire dénoyé
    TriangularWeirFree,
    // Loi de débit seuil triangulaire tronqué
    TriangularTruncWeirFree,
    // Loi de débit fente noyée (Larinier 1992)
    WeirSubmergedLarinier,
    // Loi de débit orifice noyé
    OrificeSubmerged
}

export const loiAdmissiblesOuvrages: { [key: string]: LoiDebit[] } = {
    Orifice: [
        LoiDebit.OrificeSubmerged
    ],
    SeuilRectangulaire: [
        LoiDebit.WeirCem88d, LoiDebit.WeirCem88v, LoiDebit.Cunge80, LoiDebit.WeirFree,
        LoiDebit.KIVI, LoiDebit.WeirSubmergedLarinier
    ],
    SeuilTriangulaire: [
        LoiDebit.TriangularWeirFree
    ],
    SeuilTriangulaireTrunc: [
        LoiDebit.TriangularTruncWeirFree
    ],
    VanneRectangulaire: [
        LoiDebit.GateCem88d, LoiDebit.GateCem88v, LoiDebit.Cunge80, LoiDebit.RectangularOrificeFree,
        LoiDebit.RectangularOrificeSubmerged
    ]
};

export const loiAdmissiblesCloisons: { [key: string]: LoiDebit[] } = {
    Orifice: [
        LoiDebit.OrificeSubmerged
    ],
    SeuilRectangulaire: [
        LoiDebit.KIVI, LoiDebit.WeirSubmergedLarinier
    ]
};

export const loiAdmissiblesDever: { [key: string]: LoiDebit[] } = {
    SeuilRectangulaire: [
        LoiDebit.WeirFree
    ],
    SeuilTriangulaire: [
        LoiDebit.TriangularWeirFree
    ],
    SeuilTriangulaireTrunc: [
        LoiDebit.TriangularTruncWeirFree
    ]
};

export class StructureProperties {
    /**
     * @return true si les valeurs de StructureType et LoiDebit sont compatibles
     */
    public static isCompatibleValues(struct: StructureType, loi: LoiDebit, parentNub: ParallelStructure): boolean {
        return parentNub.getLoisAdmissibles()[StructureType[struct]].includes(loi);
    }

    /**
     * @return la 1ère valeur de StructureType compatible avec la loi de débit, dans le contexte
     * du module de calcul parentNub
     */
    public static findCompatibleStructure(loi: LoiDebit, parentNub: ParallelStructure): StructureType {
        const loisAdmissibles = parentNub.getLoisAdmissibles();
        for (const st in loisAdmissibles) {
            if (loisAdmissibles.hasOwnProperty(st)) {
                const lds: LoiDebit[] = loisAdmissibles[st];
                for (const ld of lds) {
                    if (ld === loi) { return (StructureType as any)[st]; }
                }
            }
        }
        return undefined;
    }

    /**
     * trouve la 1ère valeur de LoiDebit compatible avec le type de structure
     * @param struct type de structure avec laquelle la loi de débit doit être compatible
     * @param subset si non vide, recherche la loi de débit compatible dans ce tableau; sinon prend la 1ere
     */
    public static findCompatibleLoiDebit(struct: StructureType, subset: LoiDebit[],
                                         parentNub: ParallelStructure): LoiDebit {

        const sst: string = StructureType[struct];
        const lois = parentNub.getLoisAdmissibles();

        for (const ld of lois[sst]) {
            if (subset.length === 0 || subset.includes(ld)) {
                return ld;
            }
        }
        return undefined;
    }
}
